/* 
Parallel Fractal Demonstration - ParallelFractalEngine.c
Copyright  2000-6  Dean E. Dauger  (DED)

This code requires MacMPI_X.c and mpi.h available at 
http://exodus.physics.ucla.edu/appleseed/ 
- the AppleSeed Project at the Department of 
Physics, University of California, Los Angeles.  

This is a product of Dauger Research: http://daugerresearch.com/

****************************************************************
License:

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU General Public License is available by:
Writing to the Free Software Foundation, Inc., 
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
or accessing http://www.fsf.org/copyleft/gpl.html

****************************************************************

Version History:

2006 February 9 - Go to Universal.  Converted project and source code to 
 Xcode 2.1 and made to compile for both PowerPC and Intel, in 10 minutes!
 Then dealt with endians, bundles, plsts.... - DED

2001 October 28 - Added a mechanism to launch self onto four local nodes 
 automatically via AppleEvents to Pooch. - DED

2001 July 26 - A mechanism to order the >0 nodes to quit when node 0 
 quits. - DED

2001 May 7 - A timer added to nodes >0 to allow for final messages
 and recompiled with a new MacMPI_X.c library. - DED

2001 April 13 - Carbonized for OS 9 + CarbonLib and OS X and 
 recompiled with the MacMPI_X.c library. - DED

2000 June 26 - With minor substitutions in Main.c and Engine.c, recompiled 
	in CodeWarrior Pro 5 rather than CW5.  Note that, for best performance, 
	set the Global Optimizations Level to one to approach the performance 
	CW5 delivered.  Any higher setting results in slower performing code!  
	(This is why I was so reluctant to produce a "Pro" version.)  - DED

2000 March 23 - Added better comments to Engine.c and Engine.h  - DED
2000 March 9 - First Gnu Public License release

1999 January through March - Revisions
	Originally, all nodes the floating-point results of their calculations 
	to node 0, which then did all the conversions to pixels.  Later, I had 
	the other nodes convert their floating-point results to pixels, which 
	allowed the message sizes to be cut in half.  The difficulty was that 
	the distribution of pixels in the GWorld is not guaranteed to be 
	continguous, and that distribution might not be the same on different 
	processors (and it would be silly to send data with bubbles in it), 
	so I had to set up a way to write the pixels contiguously in the each 
	node's GWorld (except for node 0) and have node 0 unpack the rows of 
	pixels into the format appropriate for node 0's GWorld.  The #ifdef 
	SendPixelData switches the code between these two cases.  
1998 December 2 - Creation in Metrowerks CodeWarrior 5 (circa 1994)
*/

#ifdef __MWERKS__
#include <fp.h>
#include <QDOffscreen.h>
#include <ColorPicker.h>
#endif

#include "ParallelFractalEngine.h"
#include "mpi.h"

/* The following are a collection of utility routines */

void ConCat (unsigned char *inS1,unsigned char *inS2,unsigned char *outS);	//Concatenates two pascal strings into a third
void ConCat (unsigned char *inS1,unsigned char *inS2,unsigned char *outS)	//Concatenates two pascal strings into a third
	//	unsigned char *inS1,*inS2,*outS;
	{ short lng=0; 
	for(;lng<=inS1[0];lng++) outS[lng]=inS1[lng];
	for(;lng-inS1[0]<=inS2[0];lng++) outS[lng]=inS2[lng-inS1[0]];
	lng--; outS[0]=lng;
	}


/* Complex2PhaseColor converts a Complex number into RGB,
	mapping red to real and cyan to negative real, and the 
	magnitude squared to the brightness (approximately).  
	It first creates a lookup table of RGB tables based on HSV2RGB.  
	This mapping is appropriate for visualization of quantum 
	mechanics.  
	
   For the Parallel Fractal Demo, I'm simply using the 512 entry 
	lookup table of packed 24-bit RGB values evenly distributed
	in hue it generates.   */
 
#define ColorReferenceVal	0xC000L		//=0.75*0x10000
	/*	=0.75*65536=0x0.C*0x10000	
	This describes the brightness of the pixel	*/

//#define maxSmallFract  0x0000ffffL
Handle phaseColorHandle=nil; 
long *phaseColorArray=nil; 
long Complex2PhaseColor(Complex *in, long *outPhaseOnlyColor, myReal *magSqOut); 
//Converts a Complex number into a phase color representation, compacted into 0x00rrggbb
long Complex2PhaseColor(Complex *in, long *outPhaseOnlyColor, myReal *magSqOut)	
	{long out=0; 
	if (!phaseColorHandle) {	long tli; 
		phaseColorHandle=NewHandle(514L<<2); //Allocating room for 514 longs
		MoveHHi(phaseColorHandle); HLock(phaseColorHandle); 
		phaseColorArray=(long*)*phaseColorHandle; 
		for(tli=514; tli--; ) {
			HSVColor tHSV={0,kMaximumSmallFract,kMaximumSmallFract}; //was MaxSmallFract
			RGBColor tRGB; 
			tHSV.hue=kMaximumSmallFract*((tli<256)?tli+256:tli-256)>>9; 
			HSV2RGB(&tHSV,&tRGB); 
			phaseColorArray[tli]=EndianS32_NtoB(
				((long)tRGB.red&0xff00)<<8	
				|	((long)tRGB.green&0xff00)
				|	((long)tRGB.blue)>>8
				); 
			}
		phaseColorArray+=256; 
		}
	if (phaseColorHandle) {
		myReal magSq=in->a*in->a; 
		myReal p=CPhase((*in))*invPi; //Convert [-,] to [-1,1]
		long tli, br; 
		magSq+=in->b*in->b; 
		tli=p*256; 
		br=magSq*ColorReferenceVal; 
		out=phaseColorArray[tli]; 
		if (magSqOut) *magSqOut=magSq; 
		if (outPhaseOnlyColor) *outPhaseOnlyColor=out; 
		
		tli=br*((Byte*)&out)[1]; 
		((Byte*)&out)[1]=(tli&0xff000000)?0xff:(tli>>16); 
		tli=br*((Byte*)&out)[2]; 
		((Byte*)&out)[2]=(tli&0xff000000)?0xff:(tli>>16); 
		tli=br*((Byte*)&out)[3]; 
		((Byte*)&out)[3]=(tli&0xff000000)?0xff:(tli>>16); 
		
		}
	
	return out; 
	}
	
/* SetupRenderGW sets up a 24-bit offscreen GWorld, referenced by the renderGW 
	global, of a size specified in renderGWRect, and saves miscellaneous 
	information about the GWorld's pixmap useful for direct access to the 
	GWorld's pixels.  Other routines write to this offscreen pixmap, which is 
	then dumped to the main window.  */
	
Rect renderGWRect={0,0,ySize,xSize}; //Thie default size of the offscreen GWorld
GWorldPtr renderGW=nil; 
PixMapHandle renderGWPM=nil; long bAddrGWPM=0, rBytesGWPM=0; 
long SetupRenderGW(void); 
long SetupRenderGW()
	{OSErr err; long out=0; 
	GDHandle currDevice; CGrafPtr currPort; 
	if (renderGW) DisposeGWorld(renderGW); renderGW=nil; 
	GetGWorld(&currPort, &currDevice); 
	ParamText("\pNo memory for GWorld",nil,nil,nil); 
	if (!(err=NewGWorld(&renderGW,32,&renderGWRect,nil,nil,(GWorldFlags)0)))	{
		SetGWorld(renderGW,nil); 
		renderGWPM=GetPortPixMap(renderGW); //<QuickDraw.h>
		LockPixels(renderGWPM); 
		EraseRect(&renderGWRect); 
		bAddrGWPM=(long)(*renderGWPM)->baseAddr; 
		rBytesGWPM=(*renderGWPM)->rowBytes&0x3FFF;
		}	
	else {	out=err; StopAlert(128,nil); } 
	SetGWorld(currPort, currDevice); 
	
	return out; 
	}

/* A few globals needed for the other routines */ 

extern EventRecord theEvent; //Everyone reads and writes to the same EventRecord
extern Boolean staying; //The final signal for truly quitting the application
extern WindowPtr previewWindow; //WindowPtr to the main window
extern Str255 tStr; //Scratch space string

void DumpRectToWindow(Rect *pRect); 
void DumpRectToWindow(Rect *pRect)
	{//Only if we have a GWorld
	if (renderGW) {//Dump the contents of that GWorld to the main window.
					ForeColor(blackColor); BackColor(whiteColor); //<QuickDraw.h>
					CopyBits(GetPortBitMapForCopyBits(renderGW), 
						GetPortBitMapForCopyBits(GetWindowPort(previewWindow)), 
						pRect, pRect, ditherCopy, nil); 
					BackColor(blackColor); ForeColor(whiteColor); 
				}
	}


Handle myDataHandle=nil; //Handle to one big block of data for all the floating-point output data
myReal *bigDataPtr=nil; //cached pointer to that block

//#define CountTotalIters		1
/* Turn on CountTotalIters to have it count how many iterations it does total.  
	I wrote this number down and put it into the code later to do the MFlop calculation. */
#ifdef CountTotalIters
unsigned long accumCount=0; 
#endif

/* ComputeLine computes the primary computation and saves its result into a 
	strip of myReal's pointed to by bigDataPtr and offset lineNo rows.  
	
	If all you want to do is replace the inner loops with your own calculation, 
	this is the routine to change.  Remember that the MFlops calculation will 
	then be incorrect.  */ 
void ComputeLine(long lineNo); 
void ComputeLine(long lineNo)
	{myReal *lineDataPtr=bigDataPtr+(lineNo*xSize); 
	long tx=xSize; 
	myReal cb=centerB+((ySize>>1)-lineNo)*radiusA/(xSize>>1); 
#ifdef CountTotalIters
	unsigned long lineAccumCount=0; 
#endif
	while (tx--) {
//#define testfunction	1		/* for the dummy test function */
//#define useComplexStruct	1	/* a version that used the Complex struct, for testing */
#ifdef testfunction
		lineDataPtr[tx]=(sin(lineNo*0.05)+cos(tx*0.02))*32+16; 
#else
#ifdef useComplexStruct
		Complex z, c; 
		myReal value;
		long count=0; 
		short looping=1; 
		z.a=0; z.b=0; 
		c.a=centerA+(tx-(xSize>>1))*radiusA/(xSize>>1); c.b=cb; 
				
		while (looping) {
			myReal tr; 
			count++; 
			CMultNAdd(z, z, z, c); 
			
			if (count>MaxCount) {
				looping=false; 
				value=0; 
				}
			else if ((tr=CMagSq(z))>256) {
				looping=false; 
				value=count+4;  
				
				}
			}
		
		lineDataPtr[tx]=value; 
#else
/* This has the optimized fractal iteration code.  What's nice about this 
	is that it does a lot of work, then saves one myReal, over and over.  */ 
		myReal za, zb, ca, sqrmaxrad=256; 
		long count=0; 
		short looping=1; 
		za=0; zb=0; 
		ca=centerA+(tx-(xSize>>1))*radiusA/(xSize>>1); 
		
		while (looping) {//12 flops per iteration (do a Disassemble)
			myReal tr; 
			count++; 
			tr=ca+za*za-zb*zb; //4
			zb=cb+zb*za+zb*za; //4
			za=tr;
			
			tr=tr*tr+zb*zb; //3
			if (count>MaxCount) {
				looping=false; 
				lineDataPtr[tx]=0; 
				}
			else if (tr>sqrmaxrad) {//1
				looping=false; 
				lineDataPtr[tx]=count+4; 
				//There was a good reason why the output was floating-point
				}
 			}
		
#endif

#ifdef CountTotalIters
			lineAccumCount+=count; 
#endif

#endif
		}
#ifdef CountTotalIters
	accumCount+=lineAccumCount; 
#endif
	}

/* DrawOffscreenLine translates a strip of myReal's into pixel values it 
	writes into a strip of pixels in the offscreen pixmap.  Each strip of 
	pixels is offset by rBytesGWPM, saved by SetupRenderGW.  
	
	If you want to change how the calculation output maps to pixels, 
	this is the routine to change.  */
void DrawOffscreenLine(long lineNo); 
void DrawOffscreenLine(long lineNo) 
	{myReal *lineDataPtr=bigDataPtr+(lineNo*xSize); 
	long *pixelDataPtr=(long *)(bAddrGWPM+(lineNo*rBytesGWPM)); 
	long tx=xSize; 
	while (tx--) {
		long colorIndex=ColorMapMultiple*lineDataPtr[tx]; 
//#define useGrayscale
#ifdef useGrayscale		/* Doesn't need an external color map.  Not as pretty, though. */
		colorIndex&=0xffL; 
		pixelDataPtr[tx]=EndianS32_NtoB((colorIndex<<16)|(colorIndex<<8)|colorIndex); 
#else
		if (colorIndex) {
			colorIndex&=0x1ffL; 
			pixelDataPtr[tx]=phaseColorArray?phaseColorArray[colorIndex-256]:(colorIndex&0xffL); 
			}
		else pixelDataPtr[tx]=0; 
#endif
		}
	}


Handle mpidataHandle=nil; //A handle to an array of MPI_Requests and related structures
MPI_Request *mpireqP=nil; // and a cached pointer to the MPI_Requests
long *myProcFlagsP=nil; // and a cached pointer to miscellaneous other flags

long computeState=0; /* This code operates in a state space: This number determines 
	what state the ComputeTask routine is in.  ComputeTask has deterministic ways of 
	behaving depending on which state it's in, and can jump only to specific other states.   
	This idea is inspired by Quantum Mechanics.  */
long linesPerCall=1; /* This number determines how many lines ComputeTask will compute
	each time it is called.  The purpose is so that the code can compute more than one 
	line per main event loop, but it shouldn't be so much that the code becomes 
	unresponsive to the user.  ComputeTask automatically regulates this value to 
	maintain a reasonable comprimise. */

long elapsedTime=0; //for timing of the overall calculation
int numprocs, myid; //Determined by MPI, if present
short mpiActive=0; //a flag whether or not MPI is active

/* This routine is a little calculator, accessible by all nodes, that lets an 
	individual node now where the top and bottom of its work is.  This calculation 
	is here in one central location, so that any adjustments to the partitioning 
	is in one spot.  
	
	If you'd like to play with the paritioning of the problem between nodes, 
	I'd recommend starting here, but some more advanced partitioning techinques 
	may require substantial changes to ComputeTask as well.  */
inline long StartLineNo(long procNum);
inline long StartLineNo(long procNum)
	{
	return numprocs?(ySize*procNum/numprocs):procNum?ySize:0; 
	}

/* The "intelligence" of the Engine code.  ComputeTask is the main coordinator and 
	regulator of initialization, computation, message-passing, user feedback, 
	and clean up.  Code particular to parallel computing is contained here.  */
short ComputeTask(void);
short ComputeTask(void)
	{short out=false; 
	
	if (!computeState) {//Initial state 0, indicating it's time for Allocation
		{Complex tC={0,0}; //Initialize the RGB lookup table
		Complex2PhaseColor(&tC, nil, nil); 
		}
		//GWorld
		SetupRenderGW(); 
		//Data
		myDataHandle=NewHandle(ySize*xSize*sizeof(myReal)); 
		if (myDataHandle) {
			short fref, vRef; 
			long dirID; 
			FSSpec nodeListFS;
			Str255 nStr; 
			MoveHHi(myDataHandle); 
			HLock(myDataHandle); 
			bigDataPtr=(myReal *)*myDataHandle; 
			
			{myReal *tp=bigDataPtr; long tli=ySize*xSize; 
			while(tli--) tp[tli]=0; //clear it
			}
			
#ifndef __MWERKS__
	if (1) {// What if I'm a bundle?
			CFBundleRef mainBundle=CFBundleGetMainBundle(); 
			if (mainBundle) {
				CFURLRef mainURL=CFBundleCopyBundleURL(mainBundle); 
				if (mainURL) {
					FSRef myFSRef; 
					FSSpec myFS; 
					OSErr err; 
					char path[512]=""; 
					if (CFURLGetFSRef(mainURL, &myFSRef)) {
						err=FSGetCatalogInfo(&myFSRef, kFSCatInfoNone, nil, nil, &myFS, nil); 
						if (noErr==err) {
							HSetVol(NULL, myFS.vRefNum, myFS.parID); 
							}
						err=FSRefMakePath(&myFSRef, (unsigned char*)path, 510); 
						if (noErr==err) {
							char *lastSlash, *tp; //clip the executable name
							for(lastSlash=tp=path; tp=strchr(tp, '/'); lastSlash=tp++) ;
							lastSlash[1]=0; 
							chdir(path); 
							}
						}
					CFRelease(mainURL); 
					}
				
				}
			}
#endif		
			
			HGetVol(nil, &vRef, &dirID); 
			//MacMPI
			SetWTitle(previewWindow, "\pCalling MPI_Init"); 
			FSMakeFSSpec(vRef, dirID, "\pnodelist_ip", &nodeListFS); 
		//if the nodelist file exists
		if (noErr==FSpOpenDF(&nodeListFS, fsRdPerm, &fref)) {
			FSClose(fref); //then MPI_Init should have a say
			if (!MPI_Init(nil,nil)) {
				MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
				MPI_Comm_rank(MPI_COMM_WORLD,&myid);
				NumToString(myid, nStr); 
				ConCat("\pI'm proc #", nStr, tStr); 
				ConCat(tStr, "\p in a ", tStr); 
				NumToString(numprocs, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p node cluster.", tStr); 
				mpiActive=1; 
				}
			else {
				myid=0; //run single node
				numprocs=1; 
				ConCat("\pMPI_Init failed.  See http://daugerresearch.com/", "\p", tStr); 
				mpiActive=0; 
				}
			SetWTitle(previewWindow, tStr); 
			}
		else {// no nodelist, so the user probably doesn't know this code can be in parallel
			myid=0; //run single node
			numprocs=1; 
		//	ParamText("\pNodelist file not found.  See http://daugerresearch.com/ for more information.", nil, nil, nil); 
		//	StopAlert(128, nil); 
			SetWTitle(previewWindow, "\pTo run this in parallel, see http://daugerresearch.com/"); 
			}
			
		if (!myid) //if I'm #0
			{GDHandle currDevice; CGrafPtr currPort; 
			long ti; Rect tRect=renderGWRect; 
			GetGWorld(&currPort, &currDevice); 
			SetGWorld(renderGW,nil); 
			//Clear my GWorld
			BackColor(blackColor);
			ForeColor(whiteColor); 
			EraseRect(&renderGWRect); 
			tRect.bottom=0; 
			// and fill it with descriptive information about what's going to happen
			for(ti=0; ti<numprocs; ti++) {
				Rect t2Rect; 
				tRect.top=tRect.bottom; 
				tRect.bottom=StartLineNo(ti+1); 
				t2Rect=tRect; 
				InsetRect(&t2Rect, 1, 1); 
				FrameRect(&t2Rect); 
				MoveTo(tRect.left+tRect.right>>1, (tRect.top+tRect.bottom>>1)); 
				NumToString(ti, nStr); 
				
				ConCat(ti?"\pWaiting for data from node #":
					"\pDisplaying data computed at node #", nStr, tStr); 
				ConCat(tStr, "\p", tStr); 
				Move(-StringWidth(tStr)>>1, 6); 
				DrawString(tStr); 
				}
			
			SetGWorld(currPort, currDevice); 
			
			SetPort(GetWindowPort(previewWindow)); 
			SizeWindow(previewWindow, xSize, ySize, true); 
			{Rect previewWindowPortRect; 
			GetWindowPortBounds(previewWindow, &previewWindowPortRect); 
			InvalWindowRect(previewWindow, &previewWindowPortRect); //Signal to the window manager an update
			}
			SetGWorld(currPort, currDevice); 
			}
			
			mpidataHandle=nil; 
			if (!myid) {
				//Node 0: Set up receive buffers and MPI_Request structures
				long ti; 
				mpidataHandle=NewHandle(numprocs*(sizeof(MPI_Request)+sizeof(long))); 
				if (mpidataHandle) {
					MoveHHi(mpidataHandle); 
					HLock(mpidataHandle); 
					mpireqP=(MPI_Request *)*mpidataHandle; 
					myProcFlagsP=(long*)&mpireqP[numprocs]; 
					myProcFlagsP[0]=0; 
					for(ti=1; ti<numprocs; ti++) {
						myProcFlagsP[ti]=0; //Clear flags
						//Set up Irecv's 
#ifdef SendPixelData
						MPI_Irecv((void*)(bAddrGWPM+(StartLineNo(ti)*rBytesGWPM)), 
							(StartLineNo(ti+1)-StartLineNo(ti))*PackedRowBytes/*>>2*/, 
							/*MPI_LONG don't want conversion*/MPI_BYTE, ti,
							ti, MPI_COMM_WORLD, &mpireqP[ti]);
#else
						MPI_Irecv(bigDataPtr+(StartLineNo(ti)*xSize), 
							(StartLineNo(ti+1)-StartLineNo(ti))*xSize, 
							myRealMPIDatatype, ti,
							ti, MPI_COMM_WORLD, &mpireqP[ti]);
#endif
						}
					//So whenever they finish, the data will arrive
					}
				}
#ifdef SendPixelData
			else {
				//modify rowBytes so that pixel data is sent in one packet without gaps
				if (renderGW) {
					rBytesGWPM=PackedRowBytes; 
					}
				}
#endif
			
			computeState=StartLineNo(myid)+1; 
			elapsedTime=TickCount(); 
			}
		else {
			bigDataPtr=nil; 
			computeState=-1; 
			ParamText("\pUnable to allocate memory", nil, nil, nil); 
			StopAlert(128, nil); 
			}
		}
	
	if (bigDataPtr) //Only if we got allocation
	switch (computeState) {//Depending on compute state
		case 0: 
		case -1: //Something's wrong: Do nothing
			break; 
		case -2: 
			elapsedTime=TickCount()+3600; 
			computeState=-3; 
		case -3: 
			if (TickCount()>elapsedTime) 
				staying=false; 
			break; 
		default: 
			{long loop=linesPerCall, startCS=computeState; 
			long tmElapsed=TickCount(); 
			short allOtherNodesFinished=true; 
			
			{long myLastLine=/*ySize*/StartLineNo(myid+1); //my node's last line is the next node's start line
			while (loop--) //Loop up to linesPerCall times
				if (computeState<=myLastLine) {
					ComputeLine(computeState-1/* convert to zero-based */); //Calculate values
#ifndef SendPixelData
					if (!myid) 
#endif
						DrawOffscreenLine(computeState-1); //Translate into pixels
					computeState++; 
					}
				else //If we've reached my last line
					loop=0; //end the loop
			}
			
		if (!myid) {Rect tRect; //Node 0 dumps pixels
			tRect.left=renderGWRect.left; 
			tRect.right=renderGWRect.right; 
			tRect.top=startCS-1; // from the top of this pass' computation
			tRect.bottom=computeState-1; // to the bottom
			if (tRect.bottom>tRect.top) 
				DumpRectToWindow(&tRect); // to the main window
			}
			
				{//Check receive buffers
				long ti; 
				if (mpidataHandle) {//Only node 0 has mpidataHandle allocated
					for(ti=1; ti<numprocs; ti++) if (!(1&myProcFlagsP[ti])) {//If this one hasn't finished
						int flag; 
						MPI_Status stat; 
						MPI_Test(&mpireqP[ti], &flag, &stat); //test if that buffer has received
						//Has that one finished?
						if (flag) {//Yes, then plot
							Rect tRect; 
							tRect.left=renderGWRect.left; 
							tRect.right=renderGWRect.right; 
							tRect.top=StartLineNo(ti); 
							tRect.bottom=StartLineNo(ti+1); 
#ifdef SendPixelData
							if (PackedRowBytes<rBytesGWPM) 
								{long line; 
	//Move pixels around to unpack the pixels rows into what the GWorld needs
								unsigned char *tp=(unsigned char*)(bAddrGWPM+(StartLineNo(ti)*rBytesGWPM)); 
								for(line=tRect.bottom-tRect.top; line--;) {
									BlockMove(tp+PackedRowBytes*line, tp+rBytesGWPM*line, PackedRowBytes); 
									}
								}
#else
							{long line; //Translate received floating-point data into pixels
							for(line=tRect.top; line<tRect.bottom; line++) 
								DrawOffscreenLine(line); 
							}
#endif
							DumpRectToWindow(&tRect); //Dump to screen
							myProcFlagsP[ti]|=1; //Set flag so we don't call MPI_Test again
							}
						else //One MPI_Test didn't return true, so we're not done
							allOtherNodesFinished=false; 
						}
					}
				}
			
			if (computeState>/*ySize*/StartLineNo(myid+1)) 
			if (myid) {//I'm done with my data, so send it to node 0
#ifdef SendPixelData
				MPI_Send((void*)(bAddrGWPM+(StartLineNo(myid)*rBytesGWPM)), 
					(StartLineNo(myid+1)-StartLineNo(myid))*PackedRowBytes/*>>2*/, 
					/*MPI_LONG don't want conversion*/MPI_BYTE, 0,
             		myid, MPI_COMM_WORLD);
#else
				MPI_Send(bigDataPtr+(StartLineNo(myid)*xSize), 
					(StartLineNo(myid+1)-StartLineNo(myid))*xSize, 
					myRealMPIDatatype, 0,
             		myid, MPI_COMM_WORLD);
#endif
				computeState=-2;//-1; 
				//staying=false;//and quit
				//MPI_Bcast(&staying, sizeof(staying), MPI_BYTE, 0, MPI_COMM_WORLD); 
				MPI_Irecv(&staying, sizeof(staying), MPI_BYTE, 0, 0, MPI_COMM_WORLD, 
					(MPI_Request *)bigDataPtr); //a bad practice, but this is one-shot anyway
				}
			else //Node 0: I'm done with my data, but ...
			if (allOtherNodesFinished) {// are the other guys all finished?
				Str32 nStr; 
				elapsedTime=TickCount()-elapsedTime; //click the stopwatch
				//Assemble the output string for display to the user
				NumToString(elapsedTime/60, nStr); 
				ConCat("\pElapsed Time: ", nStr, tStr); 
				ConCat(tStr, "\p.", tStr); 
				NumToString((elapsedTime/6)%10, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p seconds", tStr); 
#ifdef CountTotalIters
				ConCat(tStr, "\p TotalIters=", tStr); 
				NumToString(accumCount, nStr); 
				ConCat(tStr, nStr, tStr); 
#else
#ifdef __MWERKS__
#ifdef powerc
//CountTotalIters mode used to get the 187445748 value
				{long mflops=187445748.0*12*60*(1e-5)/elapsedTime; 
				ConCat(tStr, "\p - Averaging ", tStr); 
				NumToString(mflops/10, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p.", tStr); 
				NumToString(mflops%10, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p MegaFlops", tStr); 
				}
#else
	//Ummm, I'm going to have to go down to kiloFlops for 68k machines, sorry...
				{long mflops=187445748.0*12*60*(1e-3)/elapsedTime; 
				ConCat(tStr, "\p => Averaging ", tStr); 
				NumToString(mflops/1000, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p.", tStr); 
				NumToString(mflops%1000, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p MegaFlops", tStr); 
				}
#endif
#else
//CountTotalIters mode used to get the 187445748 value
				{long mflops=187445748.0*
#ifdef __ppc__
					10	/* 2 multiples optimized away */
#endif
#ifdef __i386__
					10	/* 2 multiples optimized away */
#endif
					*60*(1e-5)/elapsedTime; 
				ConCat(tStr, "\p - Averaging ", tStr); 
				NumToString(mflops/10, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p.", tStr); 
				NumToString(mflops%10, nStr); 
				ConCat(tStr, nStr, tStr); 
				ConCat(tStr, "\p MegaFlops", tStr); 
				}
#endif
#endif
				SetWTitle(previewWindow, tStr); 
				
				SetPort(GetWindowPort(previewWindow)); 
				{Rect previewWindowPortRect; 
				GetWindowPortBounds(previewWindow, &previewWindowPortRect); 
				InvalWindowRect(previewWindow, &previewWindowPortRect); 
				}
				computeState=-1; //We're done
				}
			
			//Record time elapsed in this pass
			tmElapsed=TickCount()-tmElapsed; 
/* Adjust autoregulator.  TicksPerCall is the target number of ticks per call.
	We adjust linesPerCall up or down depending on how far off our time elapsed 
	was from TicksPerCall. */
			if (!tmElapsed) linesPerCall<<=1; //if no time, double
			else if (tmElapsed<((TicksPerCall>>1)+(TicksPerCall>>2))) linesPerCall<<=1; //75% of target, double
			else if (tmElapsed<TicksPerCall) linesPerCall++; //just below, increment
			else if (tmElapsed>TicksPerCall) linesPerCall=TicksPerCall*linesPerCall/tmElapsed; //just above, zero in
			else if (tmElapsed>(TicksPerCall<<1)) linesPerCall>>=1; //2x over, half 
			if (!linesPerCall) linesPerCall=1; //If something's wrong, set to one
			
			}
			break; 
		}
	
	return out; 
	}

/* Disposes, deallocates all structures and libraries set up by the engine, 
	and resets relevant globals */
void DisposeMyMemory(); 
void DisposeMyMemory()
	{
	if (mpiActive) MPI_Finalize();
			{short vRef; 
			long dirID; 
			FSSpec nodeListFS;
			
			HGetVol(nil, &vRef, &dirID); 
			FSMakeFSSpec(vRef, dirID, "\pnodelist_ip", &nodeListFS); 
			if (noErr==FSpDelete(&nodeListFS)) {//Files.h
				}
			}
	if (mpidataHandle) {HUnlock(mpidataHandle); DisposeHandle(mpidataHandle); } 
	mpidataHandle=nil; 
	if (phaseColorHandle) {HUnlock(phaseColorHandle); DisposeHandle(phaseColorHandle);} 
	phaseColorHandle=nil; 
	if (renderGW) DisposeGWorld(renderGW); renderGW=nil; 
	if (myDataHandle) {
		HUnlock(myDataHandle); 
		DisposeHandle(myDataHandle); 
		myDataHandle=nil; 
		}
	bigDataPtr=nil; 
	computeState=0; 
	}

/* DoTask is called once in each main event loop.  Signals are sent 
	to DoTask through its message variable.  That's all the Main code needs to know.  */
void DoTask(long message)
	{short fullUpdate=false; 
	
	switch (message) {	
		case updateEvt: //I should update the window
			fullUpdate=true; 
			break; 
		case -1: //Deallocate
			if (mpiActive) 
			if (!myid) 
				MPI_Bcast(&staying, sizeof(staying), MPI_BYTE, 0, MPI_COMM_WORLD); 
			DisposeMyMemory(); 
			break; 
		default: fullUpdate=ComputeTask(); 
			break; 
		}
	
	if (fullUpdate) {
		SetPort(GetWindowPort(previewWindow)); 
		DumpRectToWindow(&renderGWRect); 
		}
	
	}

#include "PoochCalls.h"

void AutoLaunchOntoNodes(long numNodes) 
	{
	if (!myid) 
	if (numNodes>0) {
		FSSpec myFSS; 
		OSErr err; 
		
		//end the fractal calculation, including quitting other nodes
		DoTask(-1); 
		
		HideWindow(previewWindow); 
		
		if (numNodes>1) 
		if (noErr==(err=LocateMe(&myFSS))) {//where am I?
			
			err=LaunchOntoNodes(&myFSS, numNodes, true); 
			
			if (noErr==err) {
				//it's all okay
				
				}
			else 
				switch (err) {
					case kOTNotFoundErr: 
						{Str255 tStr; 
ConCat("\pPooch could not be found.  Please run the Pooch Installer on this and other machines on your network to install Pooch.  \r\rSee http://daugerresearch.com/ for more information.", "\p", tStr); 
						ParamText(tStr,nil,nil,nil); 
						}
						StopAlert(128, nil); 						
						ParamText("\p",nil,nil,nil); 
						break; 
					default: 
						SysBeep(0); 
						break; 
					}
			
			}
		ShowWindow(previewWindow); 
		
		}
	
	}
